//
// ProxyFileInfoRetriever.cpp
//

#define CINTERFACE
#define USE_STUBLESS_PROXY

#include "stdafx.h"
#include "ProxyFileInfoRetriever.h"
#include <rpcproxy.h>
#include "..\include\detours.h"			// ### WARNING: you need to download the Detours library from MSResearch.
										// Go to: http://research.microsoft.com/sn/detours/

//
// Definitions.
//

static const GUID CLSID_PSOAInterface =
	{ 0x00020424, 0, 0, { 0xC0, 0, 0, 0, 0, 0, 0, 0x46 } };
static const GUID CLSID_PSDispatch =
	{ 0x00020420, 0, 0, { 0xC0, 0, 0, 0, 0, 0, 0, 0x46 } };
static const GUID g_guidNull =
	{ 0x0000000, 0x0000, 0x0000, { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 } };

//
// NDR class.
//

DETOUR_TRAMPOLINE( HRESULT RPC_ENTRY Trampoline_NdrDllGetClassObject( IN REFCLSID rclsid, IN REFIID riid, OUT void **ppv, IN const ProxyFileInfo **pProxyFileList, IN const CLSID *pclsid, IN CStdPSFactoryBuffer* pPSFactoryBuffer ), NdrDllGetClassObject )
HRESULT RPC_ENTRY Hooked_NdrDllGetClassObject( IN REFCLSID rclsid, IN REFIID riid, OUT void **ppv, IN const ProxyFileInfo **pProxyFileList, IN const CLSID *pclsid, IN CStdPSFactoryBuffer* pPSFactoryBuffer );

typedef HRESULT ( STDAPICALLTYPE * PFNDllGetClassObject ) ( REFCLSID rclsid, REFIID riid, LPVOID * ppv );
typedef void ( RPC_ENTRY * PFNGetProxyDllInfo )( const ProxyFileInfo*** pInfo, const CLSID ** pId );

typedef std::vector< ProxyFileInfo* >		VPProxyFileInfos;

struct SClsIdModName
{
	CLSID					clsid;
	charstring				modname;
	HMODULE					hmod;
	CHAR					szPSDllName[ MAX_PATH ];
	PFNDllGetClassObject	pfnDllGetClassObject;
	PFNGetProxyDllInfo		pfnGetProxyDllInfo;

	BOOL					getclsCalled;
	VPProxyFileInfos		vpPFIs;
};
typedef std::vector< SClsIdModName >		VClsIdModName;

class CProxyFileInfoRetriever
{
public:

	//
	// Construction.
	//

	CProxyFileInfoRetriever ()
	{
		if ( pINSTANCE )
			throw;
		else
			pINSTANCE = this;

		::InitializeCriticalSection( & csGetCs );

		::DetourFunctionWithTrampoline( (PBYTE) Trampoline_NdrDllGetClassObject, (PBYTE) Hooked_NdrDllGetClassObject );
	}

	virtual ~CProxyFileInfoRetriever ()
	{
		::DetourRemove( (PBYTE) Trampoline_NdrDllGetClassObject, (PBYTE) Hooked_NdrDllGetClassObject );

		::DeleteCriticalSection( & csGetCs );

		for ( int i=0; i<vModNames.size(); i ++ )
			::FreeLibrary( vModNames[ i ].hmod );

		pINSTANCE = NULL;
	}

	//
	// Methods.
	//

	class CGetSync
	{
	public:
		CGetSync ()
		{
			::EnterCriticalSection( & CProxyFileInfoRetriever::pINSTANCE->csGetCs );
		}
		virtual ~CGetSync ()
		{
			::LeaveCriticalSection( & CProxyFileInfoRetriever::pINSTANCE->csGetCs );
		}
	};

	BOOL Get( IN REFIID riid, OUT VPProxyFileInfos& out, OUT OPTIONAL charstring* pcsPSDllName = NULL, OUT OPTIONAL CLSID* pclsidPS = NULL )
	{
		// init params.

		out.clear ();
		if ( pcsPSDllName )
			* pcsPSDllName = "";
		if ( pclsidPS )
			::memset( pclsidPS, 0, sizeof( CLSID ) );

		// get ndr data.

		CGetSync			sync;
		// // // // // // // // // // // // // // // // // // // // // // // // // // //
		{
			BOOL			retval = FALSE;

			// Get the CLSID of the Proxy/Stub module.

			CLSID		psclsid;
			::memset( & psclsid, 0, sizeof( psclsid ) );

			HRESULT		hr;

			hr = ::CoGetPSClsid( riid, & psclsid );
			if ( SUCCEEDED( hr ) )
			{
				// Check out whether this CLSID was already encountered.

				int			i;

				SClsIdModName*		pMOD = NULL;

				for ( i=0; i<vModNames.size(); i ++ )
					if ( ::IsEqualCLSID( psclsid, vModNames[ i ].clsid ) )
					{
						pMOD = & vModNames[ i ];
						break;
					}

				// If not, get the module name from the Registry.

				if ( pMOD == NULL )
				{
					LPOLESTR		clsidOleStr = NULL;
					if ( ::StringFromCLSID( psclsid, & clsidOleStr ) == S_OK )
					{
						charstring		kname =
							"CLSID\\" + ::WideStringToCharString( widestring( clsidOleStr ) ) + "\\InprocServer32";

						LONG			res;
						HKEY			hkey;

						res = ::RegOpenKey( HKEY_CLASSES_ROOT, kname.c_str(), & hkey );
						if ( res == ERROR_SUCCESS )
						{
							DWORD		dwType = 0;
							CHAR		path[ MAX_PATH ] = "";
							DWORD		dwLen = sizeof( path );

							res = ::RegQueryValueEx( hkey, NULL, NULL, & dwType, (LPBYTE) path, & dwLen );

							if ( res == ERROR_SUCCESS && dwType == REG_EXPAND_SZ )
							{
								CHAR		szExp[ MAX_PATH ] = "";
								if ( ::ExpandEnvironmentStrings( path, szExp, sizeof( szExp ) ) == 0 )
								{
									res = ERROR_BAD_FORMAT;
								}
								else
								{
									::strcpy( path, szExp );
									dwType = REG_SZ;
								}
							}

							if ( res == ERROR_SUCCESS && dwType == REG_SZ )
							{
								// Load the module.

								HMODULE			hmod = ::LoadLibrary( path );
								if ( hmod )
								{
									PFNDllGetClassObject	pfnDllGetClassObject =
										(PFNDllGetClassObject) ::GetProcAddress( hmod, "DllGetClassObject" );
									PFNGetProxyDllInfo		pfnGetProxyDllInfo =
										(PFNGetProxyDllInfo) ::GetProcAddress( hmod, "GetProxyDllInfo" );

									if ( pfnDllGetClassObject == NULL )
									{
										::FreeLibrary( hmod );
									}
									else
									{
										SClsIdModName		item;
										item.clsid = psclsid;
										item.modname = path;
										item.hmod = hmod;
										::strcpy( item.szPSDllName, path );
										item.pfnDllGetClassObject = pfnDllGetClassObject;
										item.pfnGetProxyDllInfo = pfnGetProxyDllInfo;
										item.getclsCalled = FALSE;

										vModNames.push_back( item );
										pMOD = & vModNames[ vModNames.size() - 1 ];
									}
								}
							}

							::RegCloseKey( hkey );
						}

						::CoTaskMemFree( clsidOleStr );
					}
				}

				// If the module was found, call the GetClassObject function, with an arbitrary CLSID.

				if ( pMOD )
				{
					if ( pMOD->getclsCalled == FALSE )
					{
						pCURRMOD = pMOD;
						pCURRTHREAD = ::GetCurrentThreadId();

							// call the export function, if present.

							if ( pMOD->pfnGetProxyDllInfo )
							{
								const ProxyFileInfo**	ppPFI = NULL;
								const CLSID*			pCLSID = NULL;
								pMOD->pfnGetProxyDllInfo( & ppPFI, & pCLSID );
								if ( ppPFI && * ppPFI )
									pMOD->vpPFIs.push_back( const_cast<ProxyFileInfo*>( * ppPFI ) );
							}

							// try making a fake call to DllGetClassObject: this will trigger our NDR low-level hook.

							IClassFactory*		pFactory = NULL;
							if ( SUCCEEDED( pMOD->pfnDllGetClassObject( pMOD->clsid, IID_IClassFactory, (LPVOID*) & pFactory ) ) )
								pFactory->lpVtbl->Release ( pFactory ); // should never happen !

						pCURRMOD = NULL;
						pCURRTHREAD = 0;

						pMOD->getclsCalled = TRUE;
					}

					if ( pMOD->vpPFIs.size() )
					{
						out = pMOD->vpPFIs;
						retval = TRUE;
					}

					if ( pcsPSDllName )
						* pcsPSDllName = pMOD->szPSDllName;

					if ( pclsidPS )
						* pclsidPS = pMOD->clsid;
				}
			}

			return retval;
		}
		// // // // // // // // // // // // // // // // // // // // // // // // // // //
	}

	//
	// Instance Pointer.
	//

	static CProxyFileInfoRetriever*		pINSTANCE;

protected:

	//
	// State.
	//

	CRITICAL_SECTION			csGetCs;

	VClsIdModName				vModNames;

	static SClsIdModName*		pCURRMOD;
	static DWORD				pCURRTHREAD;

	//
	// Implementation.
	//

	VOID RPC_ENTRY NdrDllGetClassObject( IN REFCLSID rclsid, IN REFIID riid, OUT void **ppv, IN const ProxyFileInfo **pProxyFileList, IN const CLSID *pclsid, IN CStdPSFactoryBuffer* pPSFactoryBuffer )
	{
		if ( pCURRMOD && pCURRTHREAD == ::GetCurrentThreadId() && pProxyFileList )
		{
			ProxyFileInfo*		pfi = * (ProxyFileInfo**) pProxyFileList;
			if ( pfi )
			{
				BOOL			found = FALSE;

				for( int i=0; i<pCURRMOD->vpPFIs.size(); i ++ )
					if ( pCURRMOD->vpPFIs[ i ] == pfi )
					{
						found = TRUE;
						break;
					}

				if ( found == FALSE )
					pCURRMOD->vpPFIs.push_back( pfi );
			}
		}
	}

	//
	// Friends.
	//

	friend
		HRESULT RPC_ENTRY Hooked_NdrDllGetClassObject( IN REFCLSID rclsid, IN REFIID riid, OUT void **ppv, IN const ProxyFileInfo **pProxyFileList, IN const CLSID *pclsid, IN CStdPSFactoryBuffer* pPSFactoryBuffer );
	friend
		class CGetSync;
};

SClsIdModName*					CProxyFileInfoRetriever::pCURRMOD = NULL;
DWORD							CProxyFileInfoRetriever::pCURRTHREAD = 0;
CProxyFileInfoRetriever*		CProxyFileInfoRetriever::pINSTANCE = NULL;

HRESULT RPC_ENTRY Hooked_NdrDllGetClassObject( IN REFCLSID rclsid, IN REFIID riid, OUT void **ppv, IN const ProxyFileInfo **pProxyFileList, IN const CLSID *pclsid, IN CStdPSFactoryBuffer* pPSFactoryBuffer )
{
	if ( CProxyFileInfoRetriever::pINSTANCE )
		CProxyFileInfoRetriever::pINSTANCE->NdrDllGetClassObject( rclsid, riid, ppv, pProxyFileList, pclsid, pPSFactoryBuffer );

	return Trampoline_NdrDllGetClassObject( rclsid, riid, ppv, pProxyFileList, pclsid, pPSFactoryBuffer );
}

//
// Public function definitions.
//

CProxyFileInfoRetriever* CreatePfiRetrieverInstance ()
{
	return new CProxyFileInfoRetriever ();
}

VOID DeletePfiRetrieverInstance ( CProxyFileInfoRetriever* pInst )
{
	delete pInst;
}

BOOL GetServerInfo( IN CProxyFileInfoRetriever* pInst, IN REFIID riid, OUT MIDL_SERVER_INFO** ppmsiServerInfo, OUT ULONG* pulDispatchTableCount, OUT IID* piidBaseInterface, OUT BYTE** ppbTypes, OUT charstring& csPSDLL, OUT charstring& csErr )
{
	* ppmsiServerInfo = NULL;
	* pulDispatchTableCount = 0;
	::memset( piidBaseInterface, 0, sizeof( IID ) );
	* ppbTypes = NULL;
	csPSDLL = "";
	csErr = "";

	// get pfis.

	VPProxyFileInfos	vpPFIs;

	CLSID				clsidPS;
	::memset( & clsidPS, 0, sizeof( clsidPS ) );

	BOOL		bGetRes = pInst->Get( riid, vpPFIs, & csPSDLL, & clsidPS );

	if ( ::IsEqualGUID( clsidPS, CLSID_PSOAInterface ) )
	{
		csErr = "Typelib marshaler used !";
		return FALSE;
	}
	else if ( ::IsEqualGUID( clsidPS, CLSID_PSDispatch ) )
	{
		csErr = "PSDispatch marshaler used !";
		return FALSE;
	}

	if ( bGetRes == FALSE )
	{
		csErr = "Unable to get ProxyFileInfo structure from P/S dll !";
		return FALSE;
	}

	// Iterate through all the returned PFIs.

	for ( int x=0; x<vpPFIs.size(); x ++ )
	{
		ProxyFileInfo*		pfi = vpPFIs[ x ];

		// get the midl_server_info data.

		PCInterfaceStubVtblList*		pStubVtbl = NULL;

		try
		{
			pStubVtbl = const_cast<PCInterfaceStubVtblList*> ( pfi->pStubVtblList );
			if ( pStubVtbl == NULL )
			{
				csErr = "InterfaceStubVtblList not found in ProxyFileInfo !";
				continue;
			}
		}
		catch( ... )
		{
			csErr = "Access violation exception when reading InterfaceStubVtblList ptr !";
			continue;
		}

		ULONG					ulIntOrd = 0;
		CInterfaceStubVtbl*		pIntSVtbl = NULL;

		try
		{
			for( int i=0; ; i ++ )
			{
				CInterfaceStubVtbl*		pThis = pStubVtbl[ i ];
				if ( pThis == NULL )
				{
					break;
				}
				else if ( ::IsEqualGUID( * pThis->header.piid, riid ) )
				{
					ulIntOrd = i;
					pIntSVtbl = pThis;
					break;
				}
			}
		}
		catch( ... )
		{
			csErr = "Access violation exception when iterating through CInterfaceStubVtbl structures !";
			continue;
		}

		if ( pIntSVtbl == NULL )
		{
			csErr = "CInterfaceStubVtbl structure for specified interface not found in PCInterfaceStubVtblList !";
			continue;
		}

		MIDL_SERVER_INFO*		pServerInfo = NULL;
		ULONG					ulDispatchTableCount = 0;

		try
		{
			pServerInfo = const_cast<MIDL_SERVER_INFO*> ( pIntSVtbl->header.pServerInfo );
			ulDispatchTableCount = pIntSVtbl->header.DispatchTableCount;

			if ( pServerInfo == NULL )
			{
				csErr = "MIDL_SERVER_INFO pointer not found in CInterfaceStubVtbl structure !";
				continue;
			}
			else if ( ulDispatchTableCount < 3 )
			{
				csErr = "Invalid CInterfaceStubVtbl.DispatchTableCount ! (less than 3)";
				continue;
			}
		}
		catch( ... )
		{
			csErr = "Access violation exception when getting MIDL_SERVER_INFO pointer from CInterfaceStubVtbl structure !";
			continue;
		}

		if ( ::IsBadReadPtr( pServerInfo, sizeof( MIDL_SERVER_INFO ) ) != FALSE )
		{
			csErr = "MIDL_SERVER_INFO pointer falls in a non-readable memory area !";
			continue;
		}

		// try to get the types table pointer...

		try
		{
			* ppbTypes = const_cast<BYTE*>( pServerInfo->pStubDesc->pFormatTypes );
			if ( ::IsBadReadPtr( * ppbTypes, sizeof( BYTE ) ) )
				* ppbTypes = NULL;
		}
		catch( ... )
		{
		}

		// return success.

		* ppmsiServerInfo = pServerInfo;
		* pulDispatchTableCount = ulDispatchTableCount;

		try
		{
			if ( pfi->pDelegatedIIDs && pfi->pDelegatedIIDs[ ulIntOrd ] )
				* piidBaseInterface = * pfi->pDelegatedIIDs[ ulIntOrd ];
		}
		catch( ... )
		{
			* piidBaseInterface = g_guidNull;
		}

		return TRUE;
	}

	// return failure.

	return FALSE;
}

BYTE* GetMethodNdrData( IN MIDL_SERVER_INFO* pServerInfo, IN ULONG ulOrdinal, OUT charstring& csErr )
{
	csErr = "";

	if ( ulOrdinal < 3 )
		throw;

	// return the info.

	try
	{
		CONST USHORT*		pusFormatOffs = pServerInfo->FmtStringOffset;
		if ( pusFormatOffs == NULL )
		{
			csErr = "Format string offset table not found !";
			return NULL;
		}

		CONST BYTE*			pbFormatTable = pServerInfo->ProcString;
		if ( pbFormatTable == NULL )
		{
			csErr = "Format strings array pointer not found !";
			return NULL;
		}

		return const_cast<BYTE*>( & pbFormatTable[ pusFormatOffs[ ulOrdinal ] ] );
	}
	catch( ... )
	{
		csErr = "Access violation exception when getting method ndr format string pointer !";
		return NULL;
	}
}
